/* ir_tx.c : 赤外線リモコン送信 サンプルプログラム
 *
 * コンパイル例:
 *	gcc -o ir_tx ir_tx.c -lrt
 *  (clock_gettime関数を使用するためリアルタイム拡張オプションが必要)
 * 実行時オプション設定(引数設定)：
 *	sudo ./ir_tx <時間間隔を格納したファイル名> で実行
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <stdint.h>

/* Parameters & macros for RPI GPIO Control --------------------------------*/
// for mmap
#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000) /* GPIOのベースアドレス */
#define CM_GP_BASE				 (BCM2708_PERI_BASE + 0x101000) /* GPCLKレジスタのベースアドレス */

#define PAGE_SIZE (4*1024)
#define BLOCK_SIZE (4*1024)

// GPIO setup macros.
#define INP_GPIO_BIT(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))		//set 0b000
#define OUT_GPIO_BIT(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3)); \
						*(gpio+((g)/10)) |=  (1<<(((g)%10)*3))		//set 0b0000 -> 0b001
#define SET_GPIO_ALT(g,a) *(gpio+(((g)/10))) |= (((a)<=3?(a)+4:(a)==4?3:2)<<(((g)%10)*3))

// GPIO Word Access
#define GPIO_SET *(gpio+7)  // sets   bits which are 1 ignores bits which are 0
#define GPIO_CLR *(gpio+10) // clears bits which are 1 ignores bits which are 0
#define GPIO_GET *(gpio+13) // read   bits

// GPIO Bit Access
#define GPIO_SET_BIT(g) *(gpio+7) = 1<<g
#define GPIO_CLR_BIT(g) *(gpio+10) = 1<<g
#define GPIO_GET_BIT(g) (*(gpio+13) & (1<<g))

/* Parameters & macros for SHT1x device I/O Control -----------------------*/
// Assigned RPI GPIOpinNo.
#define RPI_IRTX 	4			//LEDポート (今回はGPCLK0を使うためGPIO4固定)
#define IR_FREQ		38000		//赤外線サブキャリア(パルス)の周波数(Hz)
#define RPI_OSC		19200000	//BCM2835接続の源振クロック周波数(Hz)
#define MAX_PULSE 2048 			//最大送信パルス数


#define CM_GP0CTL()	*(cmgp+28)	//CM_GP0CTLレジスタへのポインタ
#define CM_GP0DIV()	*(cmgp+29)	//CM_GP0DIVレジスタへのポインタ


char *reg_map;
// I/O access for grobal
volatile unsigned *gpio;
volatile unsigned *cmgp;

// Prototypes
uint32_t *map_io(off_t);
int open_filedata(char *, long *, int *);

//------------------------------------------------------
// Set up a memory regions to access IO_regs
//
uint32_t *map_io(off_t BASE){		//引数のベースアドレスでレジスタをマップ,そのポインタを戻す
	int mem_fd;
	char *reg_map;
   /* open /dev/mem */
   if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0) {
      printf("can't open /dev/mem \n");
      exit(-1);
   }

   /* mmap GPIO */
   reg_map = (char *)mmap(
      NULL,             	//Any adddress in our space will do
      BLOCK_SIZE,       	//Map length
      PROT_READ|PROT_WRITE,	// Enable reading & writting to mapped memory
      MAP_SHARED,       	//Shared with other processes
      mem_fd,           	//File to map
      BASE         			//Offset to REGs
   );

   close(mem_fd); 

   if (reg_map == MAP_FAILED) {
      printf("mmap error %d\n", (int)reg_map);//errno also set!
      exit(-1);
   }
   // Always use volatile pointer!
   return (uint32_t *)reg_map;
}

//------------------------------------------------------

int open_filedata(char *filename, long *wait, int *plscnt){

	FILE *fp;
	int i;
	int ret;
	char str[20];
	char *linep;
	
	//送信時間を記録したファイルをオープン
	if ((fp = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "%sのオープンに失敗しました．\n", filename);
		return -1;
	}
	i = 0;
	while(fgets(str,20,fp)){			//文字列->数値変換して配列に格納
		linep = str;
		while(1){
			if(*linep == '\n') break;
			wait[i] = atol(linep);
			i++;
			if(!(linep = strchr(linep,','))) break;
			linep++;
		}
	}
	fclose(fp);							//ファイルクローズ
	return 0;
}

//-------------------------------------- main

int main(int argc,char *argv[]){

	int ret;				//戻り値格納用
	uint32_t *map_reg;		//レジスタマップ
	uint32_t temp;
	int i;

	long wait[MAX_PULSE];		//送信データの時間間隔を格納する配列
	int plscnt;					//データ数
	struct timespec nowt, pret;	//clock_gettime関数で取得した時間を格納


	if (argc != 2 ) {			//引数チェック
    	fprintf(stderr, "引数が正しくありません. (送信データファイルを指定)\n");
        exit(-1);
	}

	gpio = map_io(GPIO_BASE);	//GPIO番地で一度マップする
	SET_GPIO_ALT(RPI_IRTX,0);	//RPI_IRTXポート(GPIO4)をALT0に設定
	ret = munmap((void *)gpio, BLOCK_SIZE);	//すぐに解放

	cmgp = map_io(CM_GP_BASE);	//解放後,GPCLK番地を再度マップ
								//以降これを使用する

	//引数で指定されたファイルをオープン
	if(open_filedata(argv[1],wait, &plscnt)) exit(-1);


//分周比設定 19.2MHz(ラズベリーパイの源発振器) /
#define PLS_DIV()	CM_GP0DIV() = 0x5A000000 | ((RPI_OSC / IR_FREQ) << 12)		
#define PLS_ON()	CM_GP0CTL() = 0x5A000000 | 0x11		//GPCLK0 発振ON
#define PLS_OFF()	CM_GP0CTL() = 0x5A000000 | 0x01 	//GPCLK0 発振OFF
#define PLS_BUSY()	while(CM_GP0CTL() & 0x80)			//CM_GP0レジスタのBUSY待ち
#define PLS_RESET()	CM_GP0CTL() = 0x5A000000 | 0x21		//GPCLK0 強制停止

#define Diftime(n, p) \
  (long)((n.tv_sec - p.tv_sec) * 1000000 + (n.tv_nsec - p.tv_nsec) / 1000)

	PLS_BUSY();											//Busyフラグチェック後
	PLS_OFF();											//発振をOFFにして設定
	PLS_BUSY();											//Busyフラグチェック後
	PLS_DIV();											//分周値を設定

	plscnt = (int)wait[0];								//先頭はデータ数
	for(i=1; i<plscnt; i=i+2){							//plscnt分くりかえし
		//----------------- ON
		PLS_ON();										//パルス発振ON

		clock_gettime(CLOCK_MONOTONIC, &nowt); 			//現時間を取得
		pret = nowt;
		while(wait[i] > Diftime(nowt, pret)){
			clock_gettime(CLOCK_MONOTONIC, &nowt);		//現時間を取得
		}

		//----------------- OFF
		PLS_RESET();									//パルス強制OFF(KILL)
		PLS_OFF();										//クロックソース,分周比再設定
		PLS_DIV();

		clock_gettime(CLOCK_MONOTONIC, &nowt);			//現時間を取得
		pret = nowt;
		while(wait[i+1] > Diftime(nowt, pret)){
			clock_gettime(CLOCK_MONOTONIC, &nowt);		//現時間を取得
		}
	}
	ret = munmap((void *)cmgp, BLOCK_SIZE);				//map解放
	return ret;

} // main
