/* swint.c : SW(クリック/長押し)入力 簡易割り込み制御 サンプルプログラム
 *
 * コンパイル例:
 *	gcc -o swint swint.c
 *
 * 実行例
 *	sudo ./swint [-d/-k/null]
 *
 * 実行時オプション設定(引数設定)：
 *		-d 	  : デーモン化
 *		-k	  : kill(デーモンプロセスを停止)
 *		(なし): daemon化せず起動
 *
 *	※実行時、ファイル:[pid_file_path]を生成し、
 *	  起動したデーモンプロセスのPIDをそこへ記録します.
 *	  (後でプロセスをkillする時に使用する)
 *	※デーモン化したプロセスは [ps aux]コマンドなどで確認できます.
 */
 
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <syslog.h>
#include <poll.h>
#include <sys/signal.h>
#include <fcntl.h>

#define FLAG_DAEMONIZE     1
#define FLAG_KILL          2

#define OFF_LONG 2000	//長押し判定閾値
#define PORT_NUM 22		//swを接続するポート番号

#define pid_file_path "swint.pid"	//プロセスIDを記録するファイル名

int main(int argc, char *argv[]){

	//daemon化用変数
	int flag = 0;		//daemon動作用フラグ
	int nochdir = 1;	//0以外:カレントディレクトリを "/" (ルート) にしない
	int noclose = 0;	//0    :標準入力/標準出力/標準エラーをクローズ
	FILE *pid_fd;		//プロセスID書き出し用 ファイルディスクリプタ
	pid_t pid;			//プロセスID用変数

	//sw割り込み制御用変数
	int fd;				//GPIO制御用 ファイルディスクリプタ
	int ret;
	struct pollfd pfd;	//poll関数用 ディスクリプタ
	char val[8];

	if (argc == 2 ) {			//引数チェック
		if(*argv[1] == '-'){
			switch(*(argv[1] + 1)){
				case 'd':
					flag = FLAG_DAEMONIZE;
					//printf("daemonize\n");
					break;
				case 'k':
					flag = FLAG_KILL;
                	//printf("kill\n");
					break;
				default:
					printf("OPTION: [-d]:daemonize / [-k]:kill\n");
					return -1;
			}
		}
	}
	//このオプション以外は通常実行

	//[/var/log/]にdaemon[PID]:としてログを出力
	openlog("daemon", LOG_PID, LOG_DAEMON);


	//オプション[-k]: 保存したpidを用いてデーモンプロセスを終了する
	if (flag == FLAG_KILL) {
		pid_fd = fopen(pid_file_path, "r");
		if (pid_fd != NULL) {
			fscanf(pid_fd, "%d\n", &pid);
			fclose(pid_fd);
			if (kill(pid, SIGKILL) == 0) {
				syslog(LOG_INFO, "daemon stopped.\n");
			}
		} else {
			syslog(LOG_ERR, "no daemon started.\n");
			return -1;
		}
		return 0;
	}

	//オプション[-d]: デーモンプロセスを立ち上げる
	if (flag == FLAG_DAEMONIZE) {
		if (daemon(nochdir, noclose) == -1) {
			syslog(LOG_ERR, "failed to launch daemon.\n");
			return -1;
		}
	}

	syslog(LOG_INFO, "daemon started.\n");

	// 現プロセス(デーモン)のPIDを取得して ファイルに保存(記録しておく)
	// ※あとでkillする時に必要
	pid = getpid();
	pid_fd = fopen(pid_file_path, "w+");
	if (pid_fd != NULL) {
		fprintf(pid_fd, "%d\n", pid);
		fclose(pid_fd);
	} else {
		syslog(LOG_ERR, "failed to record process id to file.\n");
		return -1;
	}

//--------------- ここからSW入力の割り込み処理(daemon化する部分)

	//(SWを接続した)GPIOポートの入力→割り込みエッジ設定

	//ポート初期化 22番ポートをSW入力に使用
	fd = open("/sys/class/gpio/export", O_WRONLY);
	write(fd, "22", 2);
	close(fd);

	//入力に設定
	fd = open("/sys/class/gpio/gpio22/direction", O_WRONLY);
	write(fd, "in", 2);
	close(fd);

	//割り込みイベント設定
	fd = open("/sys/class/gpio/gpio22/edge", O_WRONLY);
	write(fd, "both", 4);
	close(fd);

	//入力でvalue(ポート)をオープン
	fd = open("/sys/class/gpio/gpio22/value", O_RDONLY);
	pfd.fd = fd;
	pfd.events = POLLPRI;
	read(fd, val ,sizeof(val));		//ダミーリードしておく


	//キー入力割り込みイベント監視 ループ
	while(1){
		do{
			//0(SW=OFF) <--> 1(SW=ON) 変化待ち
			lseek(fd, 0, SEEK_SET);		// /sys/～/gpio/valueレコード巻き戻し
			ret = poll(&pfd, 1, -1);	// 変化待ち。-1=無限
			read(fd, val, sizeof(val));	// 値を取得
		}while (val[0] == '0');			//OFF→ONの状態まで待機
	
		//1(SW=ON) -> 0(SW=OFF) 変化待ち
		lseek(fd, 0, SEEK_SET);
		ret = poll(&pfd, 1, OFF_LONG);	//長押し判定閾値でタイムアウト
		read(fd, val, sizeof(val));
		if (ret<=0){
			//長押しコマンド実行 →引数 SW_LONG でmainrht.pyを実行
			if(system("/home/pi/work/mainrht.py SW_LONG")== -1){
				printf("SYSERR");
			}
			//printf("LONG\n");
			continue;					//まだSW=ONのまま
		}
		else{
			//クリックコマンド実行 →引数 SW_PLS でmainrht.pyを実行
			if(system("/home/pi/work/mainrht.py SW_PLS")== -1){
				printf("SYSERR");
			}
			//printf("SHORT\n");
			continue;					//SW=OFFで繰り返し
		}
	}
	close(fd);	// value(ポート)クローズ
	
	// I/O解放
	fd = open("/sys/class/gpio/unexport", O_WRONLY);
	write(fd, "22", 2);
	close(fd);


  return 0;
}