#! /usr/bin/python
# coding: UTF-8

# mainrht.py : python エアコン・コントローラ サンプルプログラム
# 実行コマンド(例): python mainrht.py <引数>
#
#

import os
import sys
import subprocess
import time
import csv
import datetime
import calendar
import fcntl

os.environ['PATH'] = '/home/pi/work'

conffile = '/home/pi/work/rht.conf'
keyfile = '/home/pi/work/key-file'
logpath = '/home/pi/work/log/'
gpioval_fs = '/sys/class/gpio/gpio'
measure_sht = 'measureth'

LED_RED = '23'
LED_BLUE = '24'
LED_GRN = '25'
				#SWクリック入力時(暑い!ボタン)用パラメータ
MASK_EXE_SEC = 5		#連続実行防止時間(秒)
VAL_STILL_SEC = 60		#①/②の動作切り替え時間(秒)

conf = {}

def irtxOn() :
  targs = ['ir_tx', '/home/pi/work/ac_on.dat']
  for i in range(3):			#エアコンONコマンド送出 念のため3回送信
    dex = subprocess.Popen(targs,
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         shell=False)
    time.sleep(3)
  return

def irtxOff() :
  targs = ['ir_tx', '/home/pi/work/ac_off.dat']
  for i in range(3):			#エアコンOFFコマンド送出 念のため3回送信
    dex = subprocess.Popen(targs,
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         shell=False)
    time.sleep(3)
  return

def led_ctl(PORT,VAL) :		#LED ON/OFF制御
  result = 0
  while result is 0:
    try:
      fl = open('/sys/class/gpio/gpio' + PORT + '/value' ,'w')
      fl.write(VAL)
      fl.close()
      result = 1
				#LEDポートがexportされていなければ,開いて再試行
				#(但し開きっぱなし)
    except IOError:
      with open('/sys/class/gpio/export', 'w') as fl2 :
        fl2.write(PORT)
      with open('/sys/class/gpio/gpio' + PORT + '/direction', 'w') as fl2 :
        fl2.write('out')

#------------------ CONFIGファイル読み込み関数
def openDict() :
  drise = {}
  dfall = {}
  fc = open(conffile,'r+')	#ファイルオープン
  reader = csv.reader(fc)
  for row in reader :
    if row[0] == 'ST' :		#各パラメータを辞書に保存
      conf[row[1]] = row[2]
    elif row[0] == 'RISE' :
      drise[row[1]] = row[2]
    elif row[0] == 'FALL' :
      dfall[row[1]] = row[2]
    else:
      pass
  fc.close()
  conf['RISE'] = dict(drise)	#RISE/FALLの各パラメータはCONFにネスト
  conf['FALL'] = dict(dfall)
  return

#------------------ CONFIGファイル書き出し関数 カンマ区切りで1行ずつ書き出し
def closeDict():
  drise = conf['RISE']		#ネストしたRISE/FALLの各パラメータを分解
  dfall = conf['FALL']
  del conf['RISE']		#分解した後の要素を削除
  del conf['FALL']
  wbuf = ''
  wbuf += '//------ 制御用パラメータ (辞書で管理)\n'
  for key,value in conf.items() :
    wbuf += 'ST,' + key + ',' + value + '\n'
  wbuf += '//------ 上昇(エアコンOFF) -> 下降(エアコンON)転換時のパラメータ\n'
  for key,value in drise.items() :
    wbuf += 'RISE,' + key + ',' + value + '\n'
  wbuf += '//------ 下降(エアコンON) -> 上昇(エアコンOFF)転換時のパラメータ\n'
  for key,value in dfall.items() :
    wbuf += 'FALL,' + key + ',' + value + '\n'
  wbuf += '//------ END\n'
  fc = open(conffile,'w')
  fc.write(wbuf)
  fc.close()
  return

#------------------ SWクリック入力時(暑い!ボタン)処理 [簡易学習アルゴリズム]
def cl_SW_PLS():
  nowP = conf['ACT_PHASE']
  invP = 'FALL' if (nowP == 'RISE') else 'RISE'

  pret = int(conf[nowP]['PRE_TURN_TIME'])
  dt = datetime.datetime.utcnow()
  nowt = calendar.timegm(dt.timetuple())

  if(nowt - pret <= MASK_EXE_SEC) :	#前回動作が近い場合は無効(連続実行防止)
    return 'QUIT'
  elif (nowt - pret >= int(conf[nowP]['PRE_TURN_MASKSEC'])): 
    #前回の遷移よりOFFSET_VALUE以上 → ①寒い(暑い)告知として処理
    conf[nowP]['OT_OFFSET']='OFF'       #一時的な閾値オフセットは解除
    #print 'No.1\n'
    return 'F_THCNG'                    #告知時の指数を閾値に反映
                                                        
  elif (nowt - pret <= VAL_STILL_SEC) :#前回変化より設定秒以内
    if (conf[nowP]['OT_OFFSET']=='OFF') : #一時的な閾値設定なし=前回が②でない
      #→②まだ寒い(暑い)! 告知。ワンタイムオフセットを設定
      conf[invP]['OT_OFFSET']='ON'      #→次回は1回だけ閾値を高く(低く)して監視
      #print 'No.2\n'
      return 'F_NOTHCNG'                #今回転換時の指数は閾値に反映しない

    else:
      conf[nowP]['OT_OFFSET'] = 'OFF'	#→③ ②のキャンセル入力として処理
      #print 'No.3\n'			#今回転換時の指数は閾値に反映
      return 'F_NOTHGNG'		#次回閾値のオフセットは無し

  else :
    #print 'No.4\n'
    return 'QUIT'

#------------------ 温湿度測定→判定/モード転換→ログ出力処理(関数)
def phaseConv(mode):

  nowP = conf['ACT_PHASE']
  invP = 'FALL' if (nowP == 'RISE') else 'RISE'

  #---- 温度/湿度測定
  dex = subprocess.Popen([measure_sht],
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         shell=False)
  dstr =  ''.join(dex.stdout.readlines())
  if (dstr.find('Tmp:')):
    #---- 温湿度取得エラー, 必要に応じて処理挿入
    return 'ERR'

  #---- 温度/湿度を切り出し+ 不快指数計算→格納
  dt = round(float(dstr[4:4+5]),2)  	#温度
  dh = round(float(dstr[-6:-1]),2)	#湿度
  dd = round((0.81*dt+0.01*dh*(0.99*dt-14.3)+46.3),2) #不快指数計算
  conf['NOW_TEMP'] = str(dt)
  conf['NOW_HUM'] = str(dh)
  conf['NOW_DISCOM'] = str(dd)

  #---- 判定ロジック
  turn = 0
  if(mode == 'AUTO') :
    dth2 = float(conf[nowP]['TH_AVG'])
    if(conf[nowP]['OT_OFFSET']=='ON') :
      dth2 += float(conf[nowP]['OT_OFFSET_VAL'])
    if(conf['NMD_MODE'] == 'ON') :
      dth2 += float(conf[nowP]['NMD_OFFSET_VAL'])

    if (nowP == 'RISE') :
      if (dd > dth2) :
        turn = 1
    else : # == 'FALL'
      if (dd < dth2) :
        turn = 1

  else :   #mode != 'AUTO' ->強制転換
	turn = 1

  #print 'turn: %d' % (turn)

  #---- 閾値を変更する場合は平均値を計算→格納
  dth = float(conf[nowP]['TH_AVG'])
  if ((turn == 1) and (mode == 'AUTO') and (conf[nowP]['OT_OFFSET']=='ON')) \
      or (mode == 'F_THCNG'):
    avgcnt = float(conf[nowP]['TH_AVGCNT'])
    dth = (float(conf[nowP]['TH_AVG']) * (avgcnt - 1) + dd) / avgcnt
    dth = round(dth,2)
    conf[nowP]['TH_AVG'] = str(dth)
    #print 'chgavr!\n'

  #print 'dth: %f' % (dth)


  #---- フェーズ転換実行 ここでRISE<=>FALLが入れ替わる
  if (turn == 1) :
    conf[nowP]['OT_OFFSET'] = 'OFF'
    if (nowP == 'RISE') :
      conf['ACT_PHASE'] = 'FALL' 	#モードはここで転換される
      #print 'to FALL AC.ON'#AC.ONコマンド実行
      #LED設定 Blue=ON Red=OFF
      led_ctl(LED_RED,'0')
      led_ctl(LED_BLUE,'1')
      irtxOn()

    else : #FALL
      conf['ACT_PHASE'] = 'RISE'
      #print 'to RISE AC.OFF'#AC.OFFコマンド実行
      #LED設定 Blue=OFF Red=ON
      led_ctl(LED_RED,'1')
      led_ctl(LED_BLUE,'0')
      irtxOff()

  #---- タイムスタンプ作成
  if (turn == 1) :
    dtt = datetime.datetime.utcnow()
    conf[invP]['PRE_TURN_TIME'] = str(calendar.timegm(dtt.timetuple()))

  #---- カンマ区切りでログファイル出力
  sdate = datetime.datetime.now().strftime('%Y%m%d%H%M')
  with open(logpath + sdate[:-4] + '.log', 'a+') as fl :
    for i in (sdate[-4:],dt,dh,dd,nowP,dth) :
      fl.write(str(i) + ',')
    if (turn == 1) :
      fl.write('to_' + invP + ',')
      fl.write(mode + '\n')
    else :
      fl.write('\n')

  return

#------------------ 各コマンドライン引数ごとの処理(関数)
def comLine(arg) :
  if (arg == 'NMON') :
    conf['NMD_MODE'] = 'ON'
    return 'QUIT'

  elif (arg == 'NMOFF') :
    conf['NMD_MODE'] = 'OFF'
    return 'QUIT'        

  elif (arg == 'AUTO') :
    return 'AUTO'

  elif (arg == 'SW_LONG'):
    if (conf['ACT_DET']=='OFF') :
      conf['ACT_DET']='ON'
      conf['ACT_PHASE']='RISE'
      #print('#RED/GRN LED ON')
      led_ctl(LED_RED,'1')
      led_ctl(LED_GRN,'1')
      return 'AUTO'

    else:
      conf['ACT_DET']='OFF'
      conf['RISE']['OT_OFFSET'] = 'OFF'
      conf['FALL']['OT_OFFSET'] = 'OFF'
      #print('#RED/BLUE/GRN LED OFF')
      led_ctl(LED_RED,'0')
      led_ctl(LED_GRN,'0')
      led_ctl(LED_BLUE,'0')
      irtxOff()		      #エアコンOFFコマンド送出(強制)
      return 'QUIT'

  elif (arg == 'SW_PLS') :
    if (conf['ACT_DET']=='ON') :
      return cl_SW_PLS()
    else :
      return 'QUIT'

  else:
    print 'ERR:Invalid argument!'
    return 'INVAL'


#-----------  MAIN
if __name__ == '__main__':
  argvs = sys.argv
  argc = len(argvs)
#arglist = ['SW_LONG','SW_PLS','DETON','DETOFF','AUTO','NMON','NMOFF']

  if(argc !=2):
    print 'ERR: Called without argument!'
    sys.exit()

  fl = open(keyfile,'r+')			#キーファイル(ダミーファイルオープン)
  fcntl.flock(fl.fileno(), fcntl.LOCK_EX)	#ロックをかけて多重起動防止

  openDict()					#コンフィグファイル読み込み

  sts = comLine(argvs[1])			#コマンドライン引数解読→実行
  if (conf['ACT_DET'] == 'ON') :
    if ((sts != 'QUIT') and (sts != 'INVAL')):
      if(phaseConv(sts)== 'ERR') :
        print 'ERR'

  closeDict()					#コンフィグファイル書き出し

  fcntl.flock(fl.fileno(), fcntl.LOCK_UN)	#ロック解除
  fl.close()
  sys.exit()

